# postmailer

An Express middleware implementing the "HTTP POST mail" principle as an HTTP-to-SMTP proxy.

The principle is that all POST requests to a user's contact URL (e.g. `/~username`) should be received as messages by that user.  This module is a proxy to make this work using existing email infrastructure, by resending received messages using SMTP.

The module is an Express middleware, and accepts raw resources (e.g. text, HTML, images or other documents) or url-encoded forms.  Multipart messages/forms are not currently supported - any external resources should be referenced by URL or sent as separate messages.

## Usage

```javascript
var httpMail = require('postmailer');

app.use(httpMail('/~:username', {
	domain: 'example.com'
}));
```

The template (first argument above) is optional.  In the above example, documents POSTed to `/~foo` will be emailed to `foo@example.com`.

Unless `Content-Disposition:` is set to `attachment` in the POST request, plain-text and HTML are unchanged, and images are displayed inline in the email.  Otherwise, the document is included as an attachment.

## Options

* `options.transport` - optional but recommended: transport for `nodemailer` to use (see below), defaults to direct mailer
* `options.webForm` - optional: whether a friendly page should be displayed on GET requests.  Defaults to `true`
    * **boolean** - enable/disable (default template)
    * **string** - template file, with two parameters `{{username}}` and `{{url}}`
    * **function** - template function, with two arguments `username` and `url`
* `options.email` - optional: determines email address to forward to
    * **string** - constant email (useful for constant paths e.g. `/contact`)
    * **function** - accepts Express request object, returns email address (or `null` if not found)
* `options.domain` - optional: if `options.email` is omitted, then the `username` parameter from the template is combined with this domain to form an email address.
* `options.defaultEmail` - optional: if `options.email` is omitted, and either `options.domain` is missing or the `username` parameter is not set, this email address is used.
* `options.name` - optional: determines display name for page
    * **string** - constant name (useful for constant paths e.g. `/contact`)
    * **function** - accepts two arguments, Express request object and email address, returns name for display (or `null` if not found)
* `options.maxInline` - optional: maximum size (in bytes) of inline content, where larger documents are sent as attachments.  Defaults to `500*1024` (500kb)

```javascript
app.use(httpMail('example.com', {
	webForm: true, /* generates a contact form for GET - provide function for HTML templating function */,
	transport: transport /* must support the .sendMail() method from "nodemailer" module */
}));
```

## Transport

You can provide a custom transport for sending mail.  This is passed to `nodemailer.createTransport()`, and can therefore be either a transport

```javascript
options.transport = {
	host: 'smtp.mailer.example.com',
	port: 25,
	auth: {
		user: 'username',
		pass: 'password'
	}
}
```

If no details/transport is supplied, this module defaults to the [direct mailer](https://www.npmjs.com/package/nodemailer-direct-transport).  Direct SMTP connections (particularly from non-static IPs) are rejected by some mail services (e.g. Gmail) - if you want to send to these services, consider using a mailer such as [Mandrill](https://www.mandrill.com/) or [Mailgun](http://www.mailgun.com/).

## Forms

Forms (`application/x-www-form-urlencoded`) are supported, and will be interpreted using the same fields as `mailto:` URLs - i.e. fields from RFC 4021 and the [message header registry](http://www.iana.org/assignments/message-headers/message-headers.xhtml), plus the extra "body" field for text content.  Typically you'll only use "to", "subject" and "body".

However, using forms is discouraged.

## Future work

The biggest issue at the moment is that all the `From:`/`CC:`/etc. headers have to contain email addresses (or they are likely to be rejected by the receiving SMTP server or mail client.  However, according to the HTTP POST mail principle, we should be able to include URLs for replying.

A good way to handle this (supporting URLs for replying, even when using a conventional email client) could be to alias these URLs to email addresses (on a domain you own).  The counterpart to this module would be an SMTP-to-URL gateway, which accepted emails from certain email addresses and convert these aliased email addresses back into HTTP POST requests.
