var stream = require('stream');
var crypto = require('crypto');
var fs = require('fs');

var express = require('express');
var bodyParser = require('body-parser');
var parseLinks = require('parse-links');
var nodemailer = require('nodemailer');
var directTransport = require('nodemailer-direct-transport');
var mime = require('mime');
var parseHeader = require('parse-http-header');

// Pipe through to another stream without changes
// - otherwise the "path" property screws things up when using the request object/stream
function copyStream(source) {
	var content = new stream.Transform();
	content._transform = function (chunk, encoding, callback) {
		this.push(chunk, encoding);
		callback();
	};
	source.pipe(content);
	return content;
}

function commaSplit(string) {
	return string.match(/([^,]|"([^"]|\\.)*")*/g);
}

function htmlEscape(value) {
	return value.replace(/&/g, '&amp;').replace(/</g, '&lt;');
}

function HttpMail(templateUrl, options) {
	if (typeof templateUrl !== 'string') {
		options = templateUrl;
		templateUrl = '/';
	}
	options = options || {};
	options.domain = options.domain || 'localhost';
	options.name = options.name || function (request, email) {
		return email.replace(/@.*/, '');
	};
	options.maxInline = options.maxInline || 500*1024; // 500kb maximum inline document

	if (typeof options.urlTransform === 'object') {
		var urlTransform = options.urlTransform;
		urlTransform.prefix = (typeof urlTransform.prefix === 'string') ? urlTransform.prefix : 'url+';
		urlTransform.suffix = urlTransform.suffix || ('@' + (urlTransform.domain || options.domain));
		urlTransform.secret = urlTransform.secret || 'secret';
		urlTransform.hmac = urlTransform.hmac || 'sha256';
		options.urlTransform = function (url, headers) {
			var fields = [].concat(options.urlTransformFields || ['Subject', 'Message-Id', 'Date']);
			var obj = {'url': url, 'hmac': urlTransform.hmac};
			fields.forEach(function (field) {
				if (headers[field]) {
					obj[field] = headers[field];
				}
			});
			var string = JSON.stringify(obj);
			var base64 = (new Buffer(string, 'utf-8')).toString('base64');

			var hmac = crypto.createHmac(urlTransform.hmac, urlTransform.secret);
			hmac.update(string);
			var sig = hmac.digest('base64');

			var email = (
				(typeof urlTransform.prefix === 'function' ? urlTransform.prefix(url, headers) : urlTransform.prefix)
				+ base64.replace(/\+/g, '-').replace(/\//g, '_').replace(/=+$/, '')
				+ '.'
				+ sig.replace(/\+/g, '-').replace(/\//g, '_').replace(/=+$/, '')
				+ urlTransform.suffix);
			return email;
		};
	}

	function getEmail(request) {
		var email = (typeof options.email === 'function') ? options.email(request) : options.email;
		if (!email) {
			if (options.domain && request.params.username) {
				email = encodeURIComponent(request.params.username) + '@' + options.domain;
			} else if (options.defaultEmail) {
				email = options.defaultEmail;
			} else {
				return null;
			}
		}
		return email;
	}
	function getName(request) {
		var email = getEmail(request);
		if (!email) return null;
		return (typeof options.name === 'function') ? options.name(request, email) : options.name;
	}

	var mailTransport = nodemailer.createTransport(options.transport || directTransport());
	mailTransport.on('log', function (message) {
		console.log(message.type + ': ' + message.message);
	});

	var app = express();
	app.use(templateUrl, function (request, response, next) {
		response.header("Access-Control-Allow-Origin", "*");
		//response.header("Access-Control-Request-Method", "GET, POST, OPTIONS");
		response.header("Access-Control-Allow-Headers", "Origin, X-Requested-With, Content-Type, Accept, Link, From, Subject");
		next();
	});
	// Automatically parse forms
	app.use(templateUrl, bodyParser.urlencoded({extended: false}));
	app.post(templateUrl, function (request, response, next) {
		var email = getEmail(request), displayName = getName(request);
		if (!email || !displayName) return next();
		var contentType = request.headers['content-type'];

		var filename = 'attachment', cid;
		var inline = !(parseFloat(request.headers['content-length']) > options.maxInline);
		if (request.headers['content-disposition']) {
			var parsed = parseHeader(request.headers['content-disposition']);
			filename = parsed.filename || filename;
			inline = inline && (parsed[0] == 'inline');
		}

		var links = parseLinks(request.headers.link || '');
		var from = request.headers.from || request.headers['x-from'] || (links['mail-from'] || links['from'] || '').replace(/^mailto\:/, '');
		// Envelope sender/target
		var sender = request.headers.sender || request.headers['x-sender'] || (links['mail-sender'] || links['sender'] || '').replace(/^mailto\:/, '') || commaSplit(from)[0];
		sender = sender.replace(/^\s*"([^"\\]|\\.)*"/, '').replace(/(^[^<]*<|>[^>]*$)/g, '');
		var target = email.replace(/^\s*"([^"\\]|\\.)*"/, '').replace(/(^[^<]*<|>[^>]*$)/g, '');

		var mail = {
			from: from,
			to: [email],
			cc: [],
			subject: request.headers.subject || request.headers['mail-subject'] || request.headers['x-mail-subject'] || '',
			text: '',
			headers: {},
			envelope: { /* only send to the "to" address, don't include CC/BCC */
				from: sender,
				to: target
			}
		};
		if (request.headers['mail-cc']) {
			mail.cc = mail.cc.concat(commaSplit(request.headers['mail-cc']));
		}
		if (request.headers['cc']) {
			mail.cc = mail.cc.concat(commaSplit(request.headers['cc']));
		}

		// TODO: copy over all suitable headers from request

		// Forms - fields from mailto: URLs, which are based on RFC6068
		if (/^application\/x-www-form-urlencoded($|\s|;)/.test(contentType)) {
			// As per rfc4021, see the "mail" entries at http://www.iana.org/assignments/message-headers/message-headers.xhtml

			// Normalise to lower-case
			var form = {};
			for (var key in request.body) {
				form[key.toLowerCase()] = request.body[key];
			}

			if (form.to) {
				mail.to = mail.to.concat(commaSplit(form.to));
			}
			['cc'].forEach(function (key) {
				if (form[key]) {
					mail[key] = commaSplit(form[key]);
				}
			});
			// We don't support BCC because actually we only want to handle the stuff for ourselves!

			var mailObjectMap = {
				'subject': 'subject',
				'body': 'text',
				'text': 'text',
				'from': 'from',
				'owner': 'sender',
				'in-reply-to': 'inReplyTo',
				'reply-to': 'replyTo',
				'references': 'references',
				'message-id': 'messageId'
			};
			for (var key in mailObjectMap) {
				if (form[key]) {
					mail[mailObjectMap[key]] = form[key];
				}
			}

			var headerList = ['comments', 'accept-language', 'disposition-notification-to', 'disposition-notification-options', 'expires', 'keywords', 'language', 'list-archive', 'list-help', 'list-id', 'list-owner', 'list-post', 'list-subscribe', 'list-unsubscribe', 'message-context', 'reply-by', 'importance', 'priority', 'sensitivity', 'message-type', 'originator-return-address', 'base'];
			headerList.forEach(function (key) {
				if (form[key]) {
					mail.headers[key] = form[key];
				}
			});
		} else if (inline && /^text\/plain($|\s|;)/.test(contentType)) {
			mail.text = copyStream(request);
		} else if (inline && /^text\/html($|\s|;)/.test(contentType)) {
			delete mail.text;
			mail.html = copyStream(request);
		} else {
			if (/^image\//.test(contentType) && inline) {
				delete mail.text;
				cid = crypto.pseudoRandomBytes(16).toString('hex') + '-' + email;
				mail.html = '<img src="cid:' + cid + '" style="max-width: 100%; max-height: 100%" alt="' + filename.replace(/&/g, '&amp;').replace(/"/g, '&quot;').replace(/</g, '&lt') + '"/>';
			}
			mail.attachments = [{
				contentType: contentType,
				content: copyStream(request),
				cid: cid,
				filename: filename
			}];
		}

		if (!mail.from) {
			return response.status(400).json('"From" is required');
		}

		if (options.urlTransform) {
			function transformField(value) {
				return value.replace(/<([^<>]+\:\/\/[^<>]+)>/g, function (match, url) {
					return '<' + options.urlTransform(url, mail.headers) + '>';
				});
			}
			for (var key in mail.headers) {
				mail.headers[key] = transformField(mail.headers[key]);
			}
			mail.from = transformField(mail.from);
			mail.cc = mail.cc.map(transformField);
		}

		mailTransport.sendMail(mail, function (error, info) {
			if (error) {
				return response.status(500).json({
					code: error.code,
					error: error.message
				});
			}
			return response.json('Message accepted');
		});
	});

	var htmlFunction;
	if (options.webForm !== false) {
		if (!options.webForm || options.webForm === true) {
			options.webForm = __dirname + '/form-template.html';
		}
		if (typeof options.webForm === 'string') {
			var template;
			try {
				template = fs.readFileSync(options.webForm, {encoding: 'utf-8'});
			} catch (e) {
				template = '<B>Error</B>: could not load template';
			}
			htmlFunction = function (username, url) {
				return template.replace(/\{\{url\}\}/g, htmlEscape(url)).replace(/\{\{username\}\}/g, htmlEscape(username));
			};
		} else if (typeof options.webForm === 'function') {
			htmlFunction = options.webForm;
		} else {
			htmlFunction = function () {
				return '<B>Error</B>: invalid template argument: ' + JSON.stringify(options.webForm).replace(/&/g, '&amp;').replace(/</g, '&lt;');
			};
		}
		app.get(templateUrl, function (request, response, next) {
			response.setHeader('Content-Type', 'text/html');
			var email = getEmail(request), displayName = getName(request);
			if (!displayName || !email) return next();

			var url = request.protocol + '://' + request.get('host') + request.originalUrl;
			response.send(htmlFunction(displayName, url));
		});
	}

	return app;
}

module.exports = HttpMail;
